"""Core font related data structures"""

from __future__ import annotations

from typing import Tuple

import displayio

class BuiltinFont:
    """A font built into CircuitPython"""

    def __init__(self) -> None:
        """Creation not supported. Available fonts are defined when CircuitPython is built. See the
        `Adafruit_CircuitPython_Bitmap_Font <https://github.com/adafruit/Adafruit_CircuitPython_Bitmap_Font>`_
        library for dynamically loaded fonts."""
        ...
    bitmap: displayio.Bitmap
    """Bitmap containing all font glyphs starting with ASCII and followed by unicode. Use
    `get_glyph` in most cases. This is useful for use with `displayio.TileGrid` and
    `terminalio.Terminal`."""
    def get_bounding_box(self) -> Tuple[int, int]:
        """Returns the maximum bounds of all glyphs in the font in a tuple of two values: width, height."""
        ...
    def get_glyph(self, codepoint: int) -> Glyph:
        """Returns a `fontio.Glyph` for the given codepoint or None if no glyph is available."""
        ...

class Glyph:
    """Storage of glyph info"""

    def __init__(
        self,
        bitmap: displayio.Bitmap,
        tile_index: int,
        width: int,
        height: int,
        dx: int,
        dy: int,
        shift_x: int,
        shift_y: int,
    ) -> None:
        """Named tuple used to capture a single glyph and its attributes.

        :param bitmap: the bitmap including the glyph
        :param tile_index: the tile index within the bitmap
        :param width: the width of the glyph's bitmap
        :param height: the height of the glyph's bitmap
        :param dx: x adjustment to the bitmap's position
        :param dy: y adjustment to the bitmap's position
        :param shift_x: the x difference to the next glyph
        :param shift_y: the y difference to the next glyph"""
        ...



"""Lightweight 2d shapes for displays"""

from __future__ import annotations

from typing import List, Tuple, Union

import displayio

class Circle:
    def __init__(
        self,
        pixel_shader: Union[displayio.ColorConverter, displayio.Palette],
        radius: int,
        x: int,
        y: int,
    ) -> None:
        """Circle is positioned on screen by its center point.

        :param pixel_shader: The pixel shader that produces colors from values
        :param radius: The radius of the circle in pixels
        :param x: Initial x position of the axis.
        :param y: Initial y position of the axis."""
    radius: int
    """The radius of the circle in pixels."""

class Polygon:
    def __init__(
        self,
        pixel_shader: Union[displayio.ColorConverter, displayio.Palette],
        points: List[Tuple[int, int]],
        x: int,
        y: int,
    ) -> None:
        """Represents a closed shape by ordered vertices

        :param pixel_shader: The pixel shader that produces colors from values
        :param points: Vertices for the polygon
        :param x: Initial screen x position of the 0,0 origin in the points list.
        :param y: Initial screen y position of the 0,0 origin in the points list."""
    points: List[Tuple[int, int]]
    """Set a new look and shape for this polygon"""

class Rectangle:
    def __init__(
        self,
        pixel_shader: Union[displayio.ColorConverter, displayio.Palette],
        width: int,
        height: int,
        x: int,
        y: int,
    ) -> None:
        """Represents a rectangle by defining its bounds

        :param pixel_shader: The pixel shader that produces colors from values
        :param width: The number of pixels wide
        :param height: The number of pixels high
        :param x: Initial x position of the top left corner.
        :param y: Initial y position of the top left corner."""



"""Global Navigation Satellite System

The `gnss` module contains classes to control the GNSS and acquire positioning information."""

from __future__ import annotations

import time
from typing import List, Union

class GNSS:
    """Get updated positioning information from Global Navigation Satellite System (GNSS)

    Usage::

        import gnss
        import time

        nav = gnss.GNSS([gnss.SatelliteSystem.GPS, gnss.SatelliteSystem.GLONASS])
        last_print = time.monotonic()
        while True:
            nav.update()
            current = time.monotonic()
            if current - last_print >= 1.0:
                last_print = current
                if nav.fix is gnss.PositionFix.INVALID:
                    print("Waiting for fix...")
                    continue
                print("Latitude: {0:.6f} degrees".format(nav.latitude))
                print("Longitude: {0:.6f} degrees".format(nav.longitude))"""

    def __init__(self, system: Union[SatelliteSystem, List[SatelliteSystem]]) -> None:
        """Turn on the GNSS.

        :param system: satellite system to use"""
        ...
    def deinit(self) -> None:
        """Turn off the GNSS."""
        ...
    def update(self) -> None:
        """Update GNSS positioning information."""
        ...
    latitude: float
    """Latitude of current position in degrees (float)."""

    longitude: float
    """Longitude of current position in degrees (float)."""

    altitude: float
    """Altitude of current position in meters (float)."""

    timestamp: time.struct_time
    """Time when the position data was updated."""

    fix: PositionFix
    """Fix mode."""

class PositionFix:
    """Position fix mode"""

    def __init__(self) -> None:
        """Enum-like class to define the position fix mode."""
    INVALID: PositionFix
    """No measurement."""

    FIX_2D: PositionFix
    """2D fix."""

    FIX_3D: PositionFix
    """3D fix."""

class SatelliteSystem:
    """Satellite system type"""

    def __init__(self) -> None:
        """Enum-like class to define the satellite system type."""
    GPS: SatelliteSystem
    """Global Positioning System."""

    GLONASS: SatelliteSystem
    """GLObal NAvigation Satellite System."""

    SBAS: SatelliteSystem
    """Satellite Based Augmentation System."""

    QZSS_L1CA: SatelliteSystem
    """Quasi-Zenith Satellite System L1C/A."""

    QZSS_L1S: SatelliteSystem
    """Quasi-Zenith Satellite System L1S."""



"""Traceback Module

This module provides a standard interface to print stack traces of programs.
This is useful when you want to print stack traces under program control.

"""

from __future__ import annotations

import io
from types import TracebackType
from typing import Optional, Type

...

def format_exception(
    etype: Type[BaseException],
    value: BaseException,
    tb: TracebackType,
    limit: Optional[int] = None,
    chain: Optional[bool] = True,
) -> None:
    """Format a stack trace and the exception information.

    The arguments have the same meaning as the corresponding arguments
    to print_exception().  The return value is a list of strings, each
    ending in a newline and some containing internal newlines.  When
    these lines are concatenated and printed, exactly the same text is
    printed as does print_exception().

    .. note: Setting `chain` will have no effect as chained exceptions are not yet implemented.

    :param Type[BaseException] etype: This is ignored and inferred from the type of ``value``.
    :param BaseException value: The exception. Must be an instance of `BaseException`.
    :param TracebackType tb: The traceback object. If `None`, the traceback will not be printed.
    :param int limit: Print up to limit stack trace entries (starting from the caller’s frame) if limit is positive.
                      Otherwise, print the last ``abs(limit)`` entries. If limit is omitted or None, all entries are printed.
    :param bool chain: If `True` then chained exceptions will be printed (note: not yet implemented).

    """
    ...

def print_exception(
    etype: Type[BaseException],
    value: BaseException,
    tb: TracebackType,
    limit: Optional[int] = None,
    file: Optional[io.FileIO] = None,
    chain: Optional[bool] = True,
) -> None:

    """Prints exception information and stack trace entries.

    .. note: Setting `chain` will have no effect as chained exceptions are not yet implemented.

    :param Type[BaseException] etype: This is ignored and inferred from the type of ``value``.
    :param BaseException value: The exception. Must be an instance of `BaseException`.
    :param TracebackType tb: The traceback object. If `None`, the traceback will not be printed.
    :param int limit: Print up to limit stack trace entries (starting from the caller’s frame) if limit is positive.
                      Otherwise, print the last ``abs(limit)`` entries. If limit is omitted or None, all entries are printed.
    :param io.FileIO file: If file is omitted or `None`, the output goes to `sys.stderr`; otherwise it should be an open
                           file or file-like object to receive the output.
    :param bool chain: If `True` then chained exceptions will be printed (note: not yet implemented).

    """
    ...



"""MIDI over USB

The `usb_midi` module contains classes to transmit and receive MIDI messages over USB."""

from __future__ import annotations

from typing import Optional, Tuple, Union

from _typing import ReadableBuffer, WriteableBuffer

ports: Tuple[Union[PortIn, PortOut], ...]
"""Tuple of all MIDI ports. Each item is ether `PortIn` or `PortOut`."""

def disable() -> None:
    """Disable presenting a USB MIDI device to the host.
    The device is normally enabled by default, but on some boards with limited endpoints
    including ESP32-S2 and certain STM boards, it is disabled by default.
    Can be called in ``boot.py``, before USB is connected."""
    ...

def enable() -> None:
    """Enable presenting a USB MIDI device to the host.
    The device is enabled by default, so you do not normally need to call this function.
    Can be called in ``boot.py``, before USB is connected.

    If you enable too many devices at once, you will run out of USB endpoints.
    The number of available endpoints varies by microcontroller.
    CircuitPython will go into safe mode after running boot.py to inform you if
    not enough endpoints are available.
    """
    ...

class PortIn:
    """Receives midi commands over USB"""

    def __init__(self) -> None:
        """You cannot create an instance of `usb_midi.PortIn`.

        PortIn objects are constructed for every corresponding entry in the USB
        descriptor and added to the ``usb_midi.ports`` tuple."""
        ...
    def read(self, nbytes: Optional[int] = None) -> Optional[bytes]:
        """Read characters.  If ``nbytes`` is specified then read at most that many
        bytes. Otherwise, read everything that arrives until the connection
        times out. Providing the number of bytes expected is highly recommended
        because it will be faster.

        :return: Data read
        :rtype: bytes or None"""
        ...
    def readinto(
        self, buf: WriteableBuffer, nbytes: Optional[int] = None
    ) -> Optional[bytes]:
        """Read bytes into the ``buf``.  If ``nbytes`` is specified then read at most
        that many bytes.  Otherwise, read at most ``len(buf)`` bytes.

        :return: number of bytes read and stored into ``buf``
        :rtype: bytes or None"""
        ...

class PortOut:
    """Sends midi messages to a computer over USB"""

    def __init__(self) -> None:
        """You cannot create an instance of `usb_midi.PortOut`.

        PortOut objects are constructed for every corresponding entry in the USB
        descriptor and added to the ``usb_midi.ports`` tuple."""
    def write(self, buf: ReadableBuffer) -> Optional[int]:
        """Write the buffer of bytes to the bus.

        :return: the number of bytes written
        :rtype: int or None"""
        ...



"""Pin references and cpu functionality

The `microcontroller` module defines the pins from the perspective of the
microcontroller. See :py:mod:`board` for board-specific pin mappings."""

from __future__ import annotations

from typing import Optional

import microcontroller
from nvm import ByteArray
from watchdog import WatchDogTimer

cpu: Processor
"""CPU information and control, such as ``cpu.temperature`` and ``cpu.frequency``
(clock frequency).
This object is an instance of `microcontroller.Processor`."""

cpus: Processor
"""CPU information and control, such as ``cpus[0].temperature`` and ``cpus[1].frequency``
(clock frequency) on chips with more than 1 cpu. The index selects which cpu.
This object is an instance of `microcontroller.Processor`."""

def delay_us(delay: int) -> None:
    """Dedicated delay method used for very short delays. **Do not** do long delays
    because this stops all other functions from completing. Think of this as an empty
    ``while`` loop that runs for the specified ``(delay)`` time. If you have other
    code or peripherals (e.g audio recording) that require specific timing or
    processing while you are waiting, explore a different avenue such as using
    `time.sleep()`."""
    ...

def disable_interrupts() -> None:
    """Disable all interrupts. Be very careful, this can stall everything."""
    ...

def enable_interrupts() -> None:
    """Enable the interrupts that were enabled at the last disable."""
    ...

def on_next_reset(run_mode: microcontroller.RunMode) -> None:
    """Configure the run mode used the next time the microcontroller is reset but
    not powered down.

    :param ~microcontroller.RunMode run_mode: The next run mode"""
    ...

def reset() -> None:
    """Reset the microcontroller. After reset, the microcontroller will enter the
    run mode last set by `on_next_reset`.

    .. warning:: This may result in file system corruption when connected to a
      host computer. Be very careful when calling this! Make sure the device
      "Safely removed" on Windows or "ejected" on Mac OSX and Linux."""
    ...

nvm: Optional[ByteArray]
"""Available non-volatile memory.
This object is the sole instance of `nvm.ByteArray` when available or ``None`` otherwise.

:type: nvm.ByteArray or None"""

watchdog: Optional[WatchDogTimer]
"""Available watchdog timer.
This object is the sole instance of `watchdog.WatchDogTimer` when available or ``None`` otherwise."""

class Pin:
    """Identifies an IO pin on the microcontroller."""

    def __init__(self) -> None:
        """Identifies an IO pin on the microcontroller. They are fixed by the
        hardware so they cannot be constructed on demand. Instead, use
        :mod:`board` or :mod:`microcontroller.pin` to reference the desired pin."""
        ...

class Processor:
    """Microcontroller CPU information and control

    Usage::

       import microcontroller
       print(microcontroller.cpu.frequency)
       print(microcontroller.cpu.temperature)

       Note that on chips with more than one cpu (such as the RP2040)
       microcontroller.cpu will return the value for CPU 0.
       To get values from other CPUs use microcontroller.cpus indexed by
       the number of the desired cpu. i.e.

       print(microcontroller.cpus[0].temperature)
       print(microcontroller.cpus[1].frequency)"""

    def __init__(self) -> None:
        """You cannot create an instance of `microcontroller.Processor`.
        Use `microcontroller.cpu` to access the sole instance available."""
        ...
    frequency: int
    """The CPU operating frequency in Hertz. (read-only)"""

    reset_reason: microcontroller.ResetReason
    """The reason the microcontroller started up from reset state."""

    temperature: Optional[float]
    """The on-chip temperature, in Celsius, as a float. (read-only)

    Is `None` if the temperature is not available."""

    uid: bytearray
    """The unique id (aka serial number) of the chip as a `bytearray`. (read-only)"""

    voltage: Optional[float]
    """The input voltage to the microcontroller, as a float. (read-only)

    Is `None` if the voltage is not available."""

class ResetReason:
    """The reason the microntroller was last reset"""

    POWER_ON: object
    """The microntroller was started from power off."""

    BROWNOUT: object
    """The microntroller was reset due to too low a voltage."""

    SOFTWARE: object
    """The microntroller was reset from software."""

    DEEP_SLEEP_ALARM: object
    """The microntroller was reset for deep sleep and restarted by an alarm."""

    RESET_PIN: object
    """The microntroller was reset by a signal on its reset pin. The pin might be connected to a reset button."""

    WATCHDOG: object
    """The microcontroller was reset by its watchdog timer."""

    UNKNOWN: object
    """The microntroller restarted for an unknown reason."""

class RunMode:
    """run state of the microcontroller"""

    def __init__(self) -> None:
        """Enum-like class to define the run mode of the microcontroller and
        CircuitPython."""
    NORMAL: RunMode
    """Run CircuitPython as normal.

    :type microcontroller.RunMode:"""

    SAFE_MODE: RunMode
    """Run CircuitPython in safe mode. User code will not run and the
    file system will be writeable over USB.

    :type microcontroller.RunMode:"""

    UF2: RunMode
    """Run the uf2 bootloader.

    :type microcontroller.RunMode:"""

    BOOTLOADER: RunMode
    """Run the default bootloader.

    :type microcontroller.RunMode:"""



"""Support for individual pulse based protocols

The `pulseio` module contains classes to provide access to basic pulse IO.
Individual pulses are commonly used in infrared remotes and in DHT
temperature sensors.

All classes change hardware state and should be deinitialized when they
are no longer needed if the program continues after use. To do so, either
call :py:meth:`!deinit` or use a context manager. See
:ref:`lifetime-and-contextmanagers` for more info."""

from __future__ import annotations

from typing import Optional

import microcontroller
from _typing import ReadableBuffer

class PulseIn:
    """Measure a series of active and idle pulses. This is commonly used in infrared receivers
    and low cost temperature sensors (DHT). The pulsed signal consists of timed active and
    idle periods. Unlike PWM, there is no set duration for active and idle pairs."""

    def __init__(
        self, pin: microcontroller.Pin, maxlen: int = 2, *, idle_state: bool = False
    ) -> None:
        """Create a PulseIn object associated with the given pin. The object acts as
        a read-only sequence of pulse lengths with a given max length. When it is
        active, new pulse lengths are added to the end of the list. When there is
        no more room (len() == `maxlen`) the oldest pulse length is removed to
        make room.

        :param ~microcontroller.Pin pin: Pin to read pulses from.
        :param int maxlen: Maximum number of pulse durations to store at once
        :param bool idle_state: Idle state of the pin. At start and after `resume`
          the first recorded pulse will the opposite state from idle.

        Read a short series of pulses::

          import pulseio
          import board

          pulses = pulseio.PulseIn(board.D7)

          # Wait for an active pulse
          while len(pulses) == 0:
              pass
          # Pause while we do something with the pulses
          pulses.pause()

          # Print the pulses. pulses[0] is an active pulse unless the length
          # reached max length and idle pulses are recorded.
          print(pulses)

          # Clear the rest
          pulses.clear()

          # Resume with an 80 microsecond active pulse
          pulses.resume(80)"""
        ...
    def deinit(self) -> None:
        """Deinitialises the PulseIn and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> PulseIn:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def pause(self) -> None:
        """Pause pulse capture"""
        ...
    def resume(self, trigger_duration: int = 0) -> None:
        """Resumes pulse capture after an optional trigger pulse.

        .. warning:: Using trigger pulse with a device that drives both high and
          low signals risks a short. Make sure your device is open drain (only
          drives low) when using a trigger pulse. You most likely added a
          "pull-up" resistor to your circuit to do this.

        :param int trigger_duration: trigger pulse duration in microseconds"""
        ...
    def clear(self) -> None:
        """Clears all captured pulses"""
        ...
    def popleft(self) -> int:
        """Removes and returns the oldest read pulse."""
        ...
    maxlen: int
    """The maximum length of the PulseIn. When len() is equal to maxlen,
    it is unclear which pulses are active and which are idle."""

    paused: bool
    """True when pulse capture is paused as a result of :py:func:`pause` or an error during capture
    such as a signal that is too fast."""
    def __bool__(self) -> bool: ...
    def __len__(self) -> int:
        """Returns the number of pulse durations currently stored.

        This allows you to::

          pulses = pulseio.PulseIn(pin)
          print(len(pulses))"""
        ...
    def __getitem__(self, index: int) -> Optional[int]:
        """Returns the value at the given index or values in slice.

        This allows you to::

          pulses = pulseio.PulseIn(pin)
          print(pulses[0])"""
        ...

class PulseOut:
    """Pulse PWM "carrier" output on and off. This is commonly used in infrared remotes. The
    pulsed signal consists of timed on and off periods. Unlike PWM, there is no set duration
    for on and off pairs."""

    def __init__(
        self,
        pin: microcontroller.Pin,
        *,
        frequency: int = 38000,
        duty_cycle: int = 1 << 15
    ) -> None:
        """Create a PulseOut object associated with the given PWMout object.

        :param ~microcontroller.Pin pin: PWMOut that is set to output on the desired pin.
        :param int frequency: Carrier signal frequency in Hertz
        :param int duty_cycle: 16-bit duty cycle of carrier frequency (0 - 65536)

        For backwards compatibility, ``pin`` may be a PWMOut object used as the carrier. This
        compatibility will be removed in CircuitPython 8.0.0.

        Send a short series of pulses::

          import array
          import pulseio
          import pwmio
          import board

          # 50% duty cycle at 38kHz.
          pwm = pulseio.PulseOut(board.D13, frequency=38000, duty_cycle=32768)
          #                             on   off     on    off    on
          pulses = array.array('H', [65000, 1000, 65000, 65000, 1000])
          pulse.send(pulses)

          # Modify the array of pulses.
          pulses[0] = 200
          pulse.send(pulses)"""
        ...
    def deinit(self) -> None:
        """Deinitialises the PulseOut and releases any hardware resources for reuse."""
        ...
    def __enter__(self) -> PulseOut:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def send(self, pulses: ReadableBuffer) -> None:
        """Pulse alternating on and off durations in microseconds starting with on.
        ``pulses`` must be an `array.array` with data type 'H' for unsigned
        halfword (two bytes).

        This method waits until the whole array of pulses has been sent and
        ensures the signal is off afterwards.

        :param array.array pulses: pulse durations in microseconds"""
        ...
